/*:
 * @plugindesc マップ画面にアイテムアイコンを表示します。
 * @author F_
 * 
 * @help
 * 
 * マップ画面にアイテムアイコンを表示します。
 * 
 * 表示アイコンは所持アイテムの増減時に自動的に追加または削除されます。
 * 
 * 
 * @param PositionX
 * @type number
 * @min -9999
 * @max 9999
 * @default 0
 * @text X座標
 * @desc 表示の左端位置(px)。
 * 
 * @param PositionY
 * @type number
 * @min -9999
 * @max 9999
 * @default 0
 * @text Y座標
 * @desc 表示の上端位置(px)。
 * 
 * @param MaxColumns
 * @type number
 * @min 1
 * @max 99
 * @default 16
 * @text 最大列数
 * @desc 横方向に並べるアイコンの最大数。
 * 
 * @param MaxRows
 * @type number
 * @min 1
 * @max 99
 * @default 4
 * @text 最大行数
 * @desc 縦方向に並べるアイコンの最大数。
 * 
 * @param IconSize
 * @type number
 * @default 0
 * @text アイコンサイズ
 * @desc アイコンの表示サイズ(px)。
 * 標準のアイコンサイズを使用する場合は0に設定。
 * 
 * @param ColumnGap
 * @type number
 * @default 0
 * @text 列間隔
 * @desc アイコン同士の横方向の間隔(px)。
 * 
 * @param RowGap
 * @type number
 * @default 0
 * @text 行間隔
 * @desc アイコン同士の縦方向の間隔(px)。
 * 
 * @param HideSwitchId
 * @type switch
 * @default 0
 * @text 非表示化スイッチ
 * @desc アイコン全体を非表示にするスイッチ。
 */

"use strict";

if (!DataManager.isBattleTest()) {
	const parseInteger = s => Number.parseInt(s, 10);

	const PLUGIN_NAME = document.currentScript.src.split("/").pop().replace(/\.js$/, "");
	const PARAMS = PluginManager.parameters(PLUGIN_NAME);
	const POSITION_X = parseInteger(PARAMS["PositionX"]);
	const POSITION_Y = parseInteger(PARAMS["PositionY"]);
	const MAX_COLUMNS = parseInteger(PARAMS["MaxColumns"]);
	const MAX_ROWS = parseInteger(PARAMS["MaxRows"]);
	const RAW_ICON_SIZE = parseInteger(PARAMS["IconSize"]);
	const COLUMN_GAP = parseInteger(PARAMS["ColumnGap"]);
	const ROW_GAP = parseInteger(PARAMS["RowGap"]);
	const HIDE_SWITCH_ID = parseInteger(PARAMS["HideSwitchId"]);

	const ICON_SIZE = RAW_ICON_SIZE || Math.max(Window_Base._iconWidth, Window_Base._iconHeight);
	const COLUMN_STRIDE = ICON_SIZE + COLUMN_GAP;
	const ROW_STRIDE = ICON_SIZE + ROW_GAP;
	const BITMAP_WIDTH = COLUMN_STRIDE * MAX_COLUMNS - COLUMN_GAP;
	const BITMAP_HEIGHT = ROW_STRIDE * MAX_ROWS - ROW_GAP;
	const MAX_ICONS = MAX_COLUMNS * MAX_ROWS;

	const appendN = (array, value, count) => {
		const result = array.slice();
		for (let i = 0; i < count; i++) {
			result.push(value);
		}
		return result;
	};

	const removeN = (array, value, count) => {
		const result = [];
		for (const [index, item] of array.entries()) {
			if (item !== value || index - result.length === count) {
				result.push(item);
			}
		}
		return result;
	};

	const __base = (obj, prop) => {
		if (obj.hasOwnProperty(prop)) {
			return obj[prop];
		} else {
			const proto = Object.getPrototypeOf(obj);
			return function () { return proto[prop].apply(this, arguments); };
		}
	};

	class ItemIconSprite extends Sprite {
		initialize() {
			super.initialize(new Bitmap(BITMAP_WIDTH, BITMAP_HEIGHT));

			this._icons = [];
		}

		isSameIconList(a, b) {
			return a.length === b.length && a.every((v, i) => v === b[i]);
		}

		update() {
			super.update();

			this.updateBitmap();
			this.updateVisibility();
		}

		updateBitmap() {
			const icons = $gameParty.itemIconsOnMap().slice(0, MAX_ICONS);
			if (!this.isSameIconList(this._icons, icons)) {
				this._icons = icons;
				this.refreshBitmap(icons);
			}
		}

		refreshBitmap(icons) {
			const bitmap = this.bitmap;

			bitmap.clear();
			for (const [index, iconIndex] of icons.entries()) {
				const xi = index % MAX_COLUMNS;
				const yi = (index - xi) / MAX_COLUMNS;
				const x = COLUMN_STRIDE * xi;
				const y = ROW_STRIDE * yi;
				this.drawIcon(bitmap, iconIndex, ICON_SIZE, x, y);
			}
		}

		drawIcon(bitmap, iconIndex, iconSize, x, y) {
			const image = ImageManager.loadSystem('IconSet');
			const pw = Window_Base._iconWidth;
			const ph = Window_Base._iconHeight;
			const sx = iconIndex % 16 * pw;
			const sy = Math.floor(iconIndex / 16) * ph;
			bitmap.blt(image, sx, sy, pw, ph, x, y, iconSize, iconSize);
		}

		updateVisibility() {
			this.visible = HIDE_SWITCH_ID === 0 || !$gameSwitches.value(HIDE_SWITCH_ID);
		}
	}

	const Game_Party_initAllItems = Game_Party.prototype.initAllItems;
	Game_Party.prototype.initAllItems = function () {
		Game_Party_initAllItems.apply(this, arguments);

		this._onMapIconItems = [];
	};

	const Game_Party_gainItem = Game_Party.prototype.gainItem;
	Game_Party.prototype.gainItem = function (item, _amount, _includeEquip) {
		if (DataManager.isItem(item)) {
			const n0 = this.numItems(item);
			Game_Party_gainItem.apply(this, arguments);
			const n1 = this.numItems(item);
			this.gainItemIconOnMap(item.id, n1 - n0);
		} else {
			Game_Party_gainItem.apply(this, arguments);
		}
	};

	Game_Party.prototype.gainItemIconOnMap = function (itemId, amount) {
		if (amount > 0) {
			this._onMapIconItems = appendN(this._onMapIconItems, itemId, amount);
		} else if (amount < 0) {
			this._onMapIconItems = removeN(this._onMapIconItems, itemId, -amount);
		}
	};

	Game_Party.prototype.itemIconsOnMap = function () {
		return this._onMapIconItems.map(itemId => {
			const item = $dataItems[itemId];
			return item !== undefined && item !== null ? item.iconIndex : 0;
		});
	};

	const Spriteset_Map_createUpperLayer = __base(Spriteset_Map.prototype, "createUpperLayer");
	Spriteset_Map.prototype.createUpperLayer = function () {
		this.createItemIconSprite();

		Spriteset_Map_createUpperLayer.apply(this, arguments);
	};

	Spriteset_Map.prototype.createItemIconSprite = function () {
		const sprite = new ItemIconSprite();
		sprite.x = POSITION_X;
		sprite.y = POSITION_Y;
		this.addChild(sprite);
	};
}